/*
 *  Copyright (C) 2010-2012 Agrosense info@agrosense.eu
 * 
 *  Licensed under the Eclipse Public License - v 1.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 * 
 *           http://www.eclipse.org/legal/epl-v10.html
 * 
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package nl.cloudfarming.client.task.shape;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeListener;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.lang.model.type.NullType;
import nl.cloudfarming.client.field.model.ActivityField;
import nl.cloudfarming.client.field.model.Field;
import nl.cloudfarming.client.field.model.FieldManager;
import nl.cloudfarming.client.field.model.Shape;
import nl.cloudfarming.client.field.shape.ShapeDataObject;
import nl.cloudfarming.client.field.shape.TreatmentZoneImportCookie;
import nl.cloudfarming.client.task.field.FieldTask;
import nl.cloudfarming.client.task.field.FieldTaskManager;
import nl.cloudfarming.client.util.ProgressListener;
import org.jdesktop.swingworker.SwingWorker;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.api.project.FileOwnerQuery;
import org.netbeans.api.project.Project;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.awt.ActionRegistration;
import org.openide.util.Lookup;
import org.openide.util.NbBundle.Messages;

@ActionID(category = "Task", id = "nl.cloudfarming.client.task.shape.CreateTaskFromShapeAction")
@ActionRegistration(iconBase = "nl/cloudfarming/client/task/shape/clipboard_sign.png", displayName = "#CTL_CreateTaskFromShape")
@ActionReferences(value = {
    @ActionReference(path = "Loaders/application/x-esri-shape/Actions", position = 3000)})
@Messages({"CTL_CreateTaskFromShape=Convert to task", "CreateTaskFromShapeAction_progress_handler_msg=Converting shapes to tasks"})
public final class CreateTaskFromShapeAction implements ActionListener {

    private final List<TreatmentZoneImportCookie> cookies;
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.task.shape.CreateTaskFromShapeAction");

    public CreateTaskFromShapeAction(List<TreatmentZoneImportCookie> cookies) {
        this.cookies = cookies;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        LOGGER.log(Level.FINEST, "performing convert to task action on {0} objects", cookies.size());
        for (TreatmentZoneImportCookie cookie : cookies) {
            ShapeDataObject shapeDataObject = cookie.getShapeDataObject();
            final List<Shape> shapes = shapeDataObject.getShapes();
            final String dataObjectFileName = shapeDataObject.getPrimaryFile().getName();
            final Project project = FileOwnerQuery.getOwner(shapeDataObject.getPrimaryFile());
            final FieldManager fieldManager = Lookup.getDefault().lookup(FieldManager.class);

            //FIXME create multi process handler
            ProgressHandle progressHandle = ProgressHandleFactory.createHandle(Bundle.CreateTaskFromShapeAction_progress_handler_msg());
            //start the progressHandle and set its limit to the number of shapes
            //contained by the shapeDataObject
            progressHandle.start(shapes.size());
            SwingWorker worker;
            final PropertyChangeListener progressListener = new ProgressListener(
                    progressHandle, FieldManager.PROP_PROGRESS, FieldManager.PROP_FINISH);

            switch (shapeDataObject.getContentType()) {
                //FIXME TV: Can the content type ever be fields??
                case ShapeDataObject.CONTENT_FIELDS:
                    //if the shapeDataObject contains fields, create a FieldWorker
                    worker = new FieldWorker(fieldManager, shapes, project, progressListener);
                    LOGGER.finest("executing field worker..");
                    break;
                case ShapeDataObject.CONTENT_TREATMENT_ZONES:
                    //if the shapeDataObject contains treatment zones, create a TreatmentZoneWorker
                    worker = new TreatmentZoneWorker(fieldManager, shapes, project, progressListener, dataObjectFileName);
                    LOGGER.finest("executing treatment zone worker..");
                    break;
                default:
                    //if the shapeDataObject contains something we cannot use, create a NullWorker
                    //that will return an empty list
                    worker = new NullWorker(progressHandle);
                    break;
            }
            worker.execute();
        }
    }

    /**
     * This method creates tasks for each activityField connected to a field. It
     * gets the activityFields for each field and creates a FieldTask from that
     * activityField.
     *
     * @param List of fields that contain activityFields
     */
    private void createTasksForFields(List<Field> fields) {
        FieldTaskManager fieldTaskManager = Lookup.getDefault().lookup(FieldTaskManager.class);
        for (Field field : fields) {
            if (field.getActivityFields().isEmpty()) {
                LOGGER.log(Level.FINEST, "Field does not contain activityFields!");
            }
            //FIXME only make task for new activity fields
            for (ActivityField activityField : field.getActivityFields()) {
                LOGGER.log(Level.FINEST, "Creating new field task for activity field with name {0}", activityField.getName());
                FieldTask fieldTask = new FieldTask(activityField);
                fieldTaskManager.addFieldTask(fieldTask);
            }
        }
    }

    /**
     * Abstract class that serves as a superclass for FieldWorker and
     * TreatmentZoneWorker. It contains the fields needed by its subclasses. The
     * done method of this class is crucial and is called by every subclass of
     * this class.
     */
    private abstract class ShapeWorker extends SwingWorker {

        protected List<Shape> shapes;
        protected FieldManager fieldManager;
        protected FieldTaskManager fieldTaskManager;
        protected Project project;
        protected PropertyChangeListener progressListener;
        protected List<Field> fields;

        public ShapeWorker(FieldManager fieldManager, List<Shape> shapes, Project project, PropertyChangeListener progressListener) {
            this.fieldManager = fieldManager;
            this.shapes = shapes;
            this.project = project;
            this.progressListener = progressListener;
            fieldTaskManager = Lookup.getDefault().lookup(FieldTaskManager.class);
        }

        @Override
        protected NullType doInBackground() throws Exception {
            fields = Collections.EMPTY_LIST;
            return null;
        }

        @Override
        protected void done() {
            createTasksForFields(fields);
        }
    }

    /**
     * This class is a specific implementation of the ShapeWorker class which is
     * used to import fields into a project in the background.
     */
    private class FieldWorker extends ShapeWorker {

        public FieldWorker(FieldManager fieldManager, List<Shape> shapes, Project project, PropertyChangeListener progressListener) {
            super(fieldManager, shapes, project, progressListener);
        }

        @Override
        protected NullType doInBackground() throws Exception {
            fields = fieldManager.importFields(shapes, project, progressListener);
            return null;
        }
    }

    /**
     * This is another specific implementation of the ShapeWorker class, capable
     * of importing treatment zones into a project in the background.
     */
    private class TreatmentZoneWorker extends ShapeWorker {

        private String dataObjectFileName;

        public TreatmentZoneWorker(FieldManager fieldManager, List<Shape> shapes, Project project, PropertyChangeListener progressListener, String dataObjectFileName) {
            super(fieldManager, shapes, project, progressListener);
            this.dataObjectFileName = dataObjectFileName;
        }

        @Override
        protected NullType doInBackground() throws Exception {
            //FIXME add new activity field to the excisting 
            //field, perhaps return list of ActivityFields?
            fields = fieldManager.importTreatmentZones(shapes, project, dataObjectFileName, progressListener);
            return null;
        }
    }

    /**
     * This class is used as the default SwingWorker in case the parameter
     * passed in this action is not a usable ShapeDataObject. This class simply
     * returns an empty list when executed.
     */
    private class NullWorker extends SwingWorker {

        private final ProgressHandle handle;

        public NullWorker(ProgressHandle handle) {
            this.handle = handle;
        }

        @Override
        protected List<Field> doInBackground() throws Exception {
            handle.finish();
            return Collections.EMPTY_LIST;
        }
    }
}
